//=============================================================================
// XC_ElementAdvancer.
// Made by Higor
// Register as HUD mutator, and use it to alter positions between Tick and PostRender
// This shit works by miracle, so whatever
//=============================================================================
class XC_ElementAdvancer expands Mutator;

const SmoothFactor = 10.f;
const DebugTickConstant = 100;

var() bool bUseSmooth;
var() bool bSetLocation;
var() bool bPhysics;

struct AdvanceInfo
{
	var Actor Actor;
	var vector PrevRelative;
	var vector OldLoc;
	var vector LocDeltas[4];
	var float LocDeltaTimes[4];
	var vector Velocity;
	var vector ChosenLoc;
	var rotator Rotation;

	var Actor Base; //For future support
	var vector BaseLocation;

	var float PhysicsAdvance;
	var bool bAdvanced;
};

var XC_CompensatorChannel Channel;

var AdvanceInfo Advanced[512]; //Slot 511 is never taken (empty struct)
var int iAdvance;
var(Debug) bool bAdvanced;

event PreBeginPlay()
{
	if ( Level.NetMode != NM_Client )
		Destroy();
}

simulated event PostRender( canvas Canvas )
{
	if ( NextHUDMutator != none )
		NextHUDMutator.PostRender( Canvas);
	if ( bAdvanced )
		DeAdvancePositions();
}

function XC_ElementAdvancer Setup( XC_CompensatorChannel Other)
{
	local Pawn P;
	Channel = Other;
	NextHUDMutator = Channel.LocalPlayer.myHud.HUDMutator;
	Channel.LocalPlayer.myHUD.HUDMutator = Self;
	bHUDMutator = True;

	ForEach AllActors (class'Pawn', P)
	{
		if ( P.Role != ROLE_SimulatedProxy || StationaryPawn(P) != none )
			continue;
		RegisterAdvance( P);
	}

	return self;
}

function AdvancePositions( float DeltaTime)
{
	local int i, j;
	local vector RelativePos, FilteredRel;
	local vector Deviation;
	local Actor ToAdvance;
	local bool bOldCA;

	for ( i=0 ; i<iAdvance ; i++ )
	{
		if ( !ValidAdvance(i) )
		{
			Advanced[i--] = Advanced[--iAdvance];
			Advanced[iAdvance] = Advanced[511];
			continue;
		}
		ToAdvance = Advanced[i].Actor;
		if ( ToAdvance.bTicked )
			Advanced[i].PhysicsAdvance = 0;
		else
			Advanced[i].PhysicsAdvance = DeltaTime;

		for ( j=3 ; j>0 ; j-- )
		{
			Advanced[i].LocDeltas[j] = Advanced[i].LocDeltas[j-1];
			Advanced[i].LocDeltaTimes[j] = Advanced[i].LocDeltaTimes[j-1];
		}
		Advanced[i].LocDeltas[0] = ToAdvance.Location - Advanced[i].OldLoc;
		Advanced[i].LocDeltaTimes[0] = DeltaTime;

		Deviation = vect(0,0,0);
		if ( VSize(Advanced[i].LocDeltas[0]) < 200 )
		{
			for ( j=1 ; j<4 ; j++ )
			{
				if ( VSize(Advanced[i].LocDeltas[j]) > 200 )
					break;
				Deviation += (Advanced[i].LocDeltas[j-1] - Advanced[i].LocDeltas[j]) / Advanced[i].LocDeltaTimes[i];
			}
			Deviation *= vect(0.65,0.65,0.3) * Channel.ProjAdv / j;
		}
		Advanced[i].OldLoc = ToAdvance.Location;
		
		RelativePos = ToAdvance.Velocity * Channel.ProjAdv + Deviation;
		Advanced[i].Base = ToAdvance.Base;
		if ( ToAdvance.Base != none )
			Advanced[i].BaseLocation = ToAdvance.Base.Location;
		else
			Advanced[i].BaseLocation = vect(0,0,0);
		if ( VSize(RelativePos) < 2 )
		{
			Advanced[i].PrevRelative = RelativePos;
			continue;
		}

		//Smooth
		FilteredRel = Advanced[i].PrevRelative * (1.f - SmoothFactor * DeltaTime) + RelativePos * SmoothFactor * DeltaTime;
//		FilteredRel = RelativePos;
		if ( bAdvanced ) //There was no de-advancement, just use this function to continue to store advance data
			continue; //I should implement a recovery method here

		if ( ToAdvance.bCollideActors )
		{
			bOldCA = true;
			ToAdvance.SetCollision( false, ToAdvance.bBlockActors, ToAdvance.bBlockPlayers);
		}
		ToAdvance.MoveSmooth( FilteredRel); //The old collision hash is unstable, let's not overload it with move commands
		if ( bOldCA )
		{
			bOldCA = false;
			ToAdvance.SetCollision( true, ToAdvance.bBlockActors, ToAdvance.bBlockPlayers);
		}
		FilteredRel = ToAdvance.Location - Advanced[i].OldLoc;
		Advanced[i].bAdvanced = true;
		Advanced[i].PrevRelative = FilteredRel;
		Advanced[i].ChosenLoc = ToAdvance.Location;
	}
	bAdvanced = true;
}

function DeAdvancePositions()
{
	local int i;
	local bool bOldCA;
	local vector TargetLocation, DiffChosen;
	local Actor ToAdvance;

	for ( i=0 ; i<iAdvance ; i++ )
	{
		if ( !ValidAdvance(i) )
		{
			Advanced[i--] = Advanced[--iAdvance];
			Advanced[iAdvance] = Advanced[511];
			continue;
		}

		if ( !Advanced[i].bAdvanced )
			continue;

		ToAdvance = Advanced[i].Actor;
		DiffChosen = ToAdvance.Location - Advanced[i].ChosenLoc;
		if ( ToAdvance.bCollideActors )
		{
			bOldCA = true;
			ToAdvance.SetCollision( false, ToAdvance.bBlockActors, ToAdvance.bBlockPlayers);
		}
		TargetLocation = Advanced[i].OldLoc + DiffChosen;
		ToAdvance.SetBase( Advanced[i].Base);
		if ( ToAdvance.Base != none && (ToAdvance.Base.Location - Advanced[i].BaseLocation != vect(0,0,0)) )
			TargetLocation -= ToAdvance.Base.Location - Advanced[i].BaseLocation;
		ToAdvance.SetLocation( TargetLocation ); //The old collision hash is unstable, let's not overload it with move commands
		if ( Advanced[i].PhysicsAdvance > 0 )
		{
			if ( ToAdvance.Physics != PHYS_None )
				ToAdvance.AutonomousPhysics( Advanced[i].PhysicsAdvance);
			else
				ToAdvance.MoveSmooth( Advanced[i].Velocity * Advanced[i].PhysicsAdvance);
		}
		if ( bOldCA )
		{
			bOldCA = false;
			ToAdvance.SetCollision( true, ToAdvance.bBlockActors, ToAdvance.bBlockPlayers);
		}

		Advanced[i].bAdvanced = false;
	}

	bAdvanced = false;
}

//This function call occurs way too often.
//Keeping it as FINAL will speedup the call opcode
final function bool ValidAdvance( int i)
{
	if ( Advanced[i].Actor == none || Advanced[i].Actor.bDeleteMe )
		return false;
	if ( Projectile(Advanced[i].Actor) != none )
	{
		if ( Advanced[i].Actor.Role == ROLE_None || Advanced[i].Actor.Role == ROLE_AutonomousProxy ) //Fix SP's and players for now
			return false;
		if ( Advanced[i].Actor.Velocity == vect(0,0,0) ) //Projectile's stopped
			return false;
	}
	return true;
}

function RegisterAdvance( Actor Other)
{
	local int i;
	Advanced[iAdvance].Actor = Other;
	Advanced[iAdvance].OldLoc = Other.Location;
	if ( Pawn(Other) != none )
		Advanced[iAdvance].Rotation = Pawn(Other).ViewRotation;
	else
		Advanced[iAdvance].Rotation = Other.Rotation;
	iAdvance++;
}

defaultproperties
{
    RemoteRole=ROLE_None
}